#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Any, List, Mapping, Tuple

from airbyte_cdk import AirbyteLogger
from airbyte_cdk.models import FailureType, SyncMode
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.utils import AirbyteTracedException
from requests.exceptions import ConnectionError, RequestException, SSLError

from .auth import MissingAccessTokenError, ShopifyAuthenticator
from .scopes import ShopifyScopes
from .streams.streams import (
    AbandonedCheckouts,
    Articles,
    BalanceTransactions,
    Blogs,
    Collections,
    Collects,
    Countries,
    CustomCollections,
    CustomerAddress,
    Customers,
    CustomerSavedSearch,
    DiscountCodes,
    Disputes,
    DraftOrders,
    FulfillmentOrders,
    Fulfillments,
    InventoryItems,
    InventoryLevels,
    Locations,
    MetafieldArticles,
    MetafieldBlogs,
    MetafieldCollections,
    MetafieldCustomers,
    MetafieldDraftOrders,
    MetafieldLocations,
    MetafieldOrders,
    MetafieldPages,
    MetafieldProductImages,
    MetafieldProducts,
    MetafieldProductVariants,
    MetafieldShops,
    MetafieldSmartCollections,
    OrderRefunds,
    OrderRisks,
    Orders,
    Pages,
    PriceRules,
    ProductImages,
    Products,
    ProductsGraphQl,
    ProductVariants,
    Shop,
    SmartCollections,
    TenderTransactions,
    Transactions,
    TransactionsGraphql,
)


class ConnectionCheckTest:
    def __init__(self, config: Mapping[str, Any]) -> None:
        self.config = config
        # use `Shop` as a test stream for connection check
        self.test_stream = Shop(self.config)
        # setting `max_retries` to 0 for the stage of `check connection`,
        # because it keeps retrying for wrong shop names,
        # but it should stop immediately
        self.test_stream.max_retries = 0

    def describe_error(self, pattern: str, shop_name: str = None, details: Any = None, **kwargs) -> str:
        connection_check_errors_map: Mapping[str, Any] = {
            "connection_error": f"Connection could not be established using `Shopify Store`: {shop_name}. Make sure it's valid and try again.",
            "request_exception": f"Request was not successfull, check your `input configuation` and try again. Details: {details}",
            "index_error": f"Failed to access the Shopify store `{shop_name}`. Verify the entered Shopify store or API Key in `input configuration`.",
            "missing_token_error": "Authentication was unsuccessful. Please verify your authentication credentials or login is correct.",
            # add the other patterns and description, if needed...
        }
        return connection_check_errors_map.get(pattern)

    def test_connection(self) -> tuple[bool, str]:
        shop_name = self.config.get("shop")
        if not shop_name:
            return False, "The `Shopify Store` name is missing. Make sure it's entered and valid."

        try:
            response = list(self.test_stream.read_records(sync_mode=SyncMode.full_refresh))
            # check for the shop_id is present in the response
            shop_id = response[0].get("id")
            if shop_id is not None:
                return True, None
            else:
                return False, f"The `shop_id` is invalid: {shop_id}"
        except (SSLError, ConnectionError):
            return False, self.describe_error("connection_error", shop_name)
        except RequestException as req_error:
            return False, self.describe_error("request_exception", details=req_error)
        except IndexError:
            return False, self.describe_error("index_error", shop_name, response)
        except MissingAccessTokenError:
            return False, self.describe_error("missing_token_error")

    def get_shop_id(self) -> str:
        """
        We need to have the `shop_id` value available to have it passed elsewhere and fill-in the missing data.
        By the time this method is tiggered, we are sure we've passed the `Connection Checks` and have the `shop_id` value.
        """
        response = list(self.test_stream.read_records(sync_mode=SyncMode.full_refresh))
        if len(response) == 0:
            raise AirbyteTracedException(
                message=f"Could not find a Shopify shop with the name {self.config.get('shop', '')}. Make sure it's valid.",
                failure_type=FailureType.config_error,
            )
        shop_id = response[0].get("id")
        if shop_id:
            return shop_id
        else:
            raise Exception(f"Couldn't get `shop_id`. Actual `response`: {response}.")


class SourceShopify(AbstractSource):
    @property
    def continue_sync_on_stream_failure(self) -> bool:
        return True

    @property
    def raise_exception_on_missing_stream(self) -> bool:
        return False

    @staticmethod
    def get_shop_name(config) -> str:
        split_pattern = ".myshopify.com"
        shop_name = config.get("shop")
        return shop_name.split(split_pattern)[0] if split_pattern in shop_name else shop_name

    @staticmethod
    def format_stream_name(name) -> str:
        return "".join(x.capitalize() for x in name.split("_"))

    def check_connection(self, logger: AirbyteLogger, config: Mapping[str, Any]) -> Tuple[bool, any]:
        """
        Testing connection availability for the connector.
        """
        config["shop"] = self.get_shop_name(config)
        config["authenticator"] = ShopifyAuthenticator(config)
        return ConnectionCheckTest(config).test_connection()

    def select_transactions_stream(self, config: Mapping[str, Any]) -> Stream:
        """
        Allow the Customer to decide which API type to use when it comes to the `Transactions` stream.
        """
        should_fetch_user_id = config.get("fetch_transactions_user_id")
        if should_fetch_user_id:
            return Transactions(config)
        else:
            return TransactionsGraphql(config)

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        """
        Mapping a input config of the user input configuration as defined in the connector spec.
        Defining streams to run.
        """
        config["shop"] = self.get_shop_name(config)
        config["authenticator"] = ShopifyAuthenticator(config)
        # add `shop_id` int value
        config["shop_id"] = ConnectionCheckTest(config).get_shop_id()
        # define scopes checker
        scopes_manager: ShopifyScopes = ShopifyScopes(config)
        # get the list of the permitted streams, based on the authenticated user scopes
        permitted_streams = scopes_manager.get_permitted_streams()
        stream_instances = [
            AbandonedCheckouts(config),
            Articles(config),
            BalanceTransactions(config),
            Blogs(config),
            Collections(config),
            Collects(config),
            CustomCollections(config),
            Customers(config),
            DiscountCodes(config),
            Disputes(config),
            DraftOrders(config),
            FulfillmentOrders(config),
            Fulfillments(config),
            InventoryItems(config),
            InventoryLevels(config),
            Locations(config),
            MetafieldArticles(config),
            MetafieldBlogs(config),
            MetafieldCollections(config),
            MetafieldCustomers(config),
            MetafieldDraftOrders(config),
            MetafieldLocations(config),
            MetafieldOrders(config),
            MetafieldPages(config),
            MetafieldProductImages(config),
            MetafieldProducts(config),
            MetafieldProductVariants(config),
            MetafieldShops(config),
            MetafieldSmartCollections(config),
            OrderRefunds(config),
            OrderRisks(config),
            Orders(config),
            Pages(config),
            PriceRules(config),
            ProductImages(config),
            Products(config),
            ProductsGraphQl(config),
            ProductVariants(config),
            Shop(config),
            SmartCollections(config),
            TenderTransactions(config),
            self.select_transactions_stream(config),
            CustomerSavedSearch(config),
            CustomerAddress(config),
            Countries(config),
        ]

        return [
            stream_instance for stream_instance in stream_instances if self.format_stream_name(stream_instance.name) in permitted_streams
        ]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from typing import Optional

import sgqlc.operation

from . import schema

_schema = schema
_schema_root = _schema.shopify_schema


# the graphql api requires the query filter to be snake case even though the column returned is camel case
def _camel_to_snake(camel_case: str):
    snake_case = []
    for char in camel_case:
        if char.isupper():
            snake_case.append("_" + char.lower())
        else:
            snake_case.append(char)
    return "".join(snake_case).lstrip("_")


def get_query_products(first: int, filter_field: str, filter_value: str, next_page_token: Optional[str]):
    op = sgqlc.operation.Operation(_schema_root.query_type)
    snake_case_filter_field = _camel_to_snake(filter_field)
    products_args = {
        "first": first,
        "query": f"{snake_case_filter_field}:>'{filter_value}'" if filter_value else None,
        "after": next_page_token,
    }
    products = op.products(**products_args)
    products.nodes.id()
    products.nodes.title()
    products.nodes.updated_at()
    products.nodes.created_at()
    products.nodes.published_at()
    products.nodes.status()
    products.nodes.vendor()
    products.nodes.product_type()
    products.nodes.tags()
    products.nodes.options()
    products.nodes.options().id()
    products.nodes.options().name()
    products.nodes.options().position()
    products.nodes.options().values()
    products.nodes.handle()
    products.nodes.description()
    products.nodes.tracks_inventory()
    products.nodes.total_inventory()
    products.nodes.total_variants()
    products.nodes.online_store_url()
    products.nodes.online_store_preview_url()
    products.nodes.description_html()
    products.nodes.is_gift_card()
    products.nodes.legacy_resource_id()
    products.nodes.media_count()
    products.page_info()
    products.page_info.has_next_page()
    products.page_info.end_cursor()
    return str(op)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import re
from typing import Any, Mapping, MutableMapping, Optional, Union
from urllib.parse import parse_qsl, urlparse

import pendulum as pdm

from .exceptions import ShopifyBulkExceptions

# default end line tag
END_OF_FILE: str = "<end_of_file>"
BULK_PARENT_KEY: str = "__parentId"


class BulkTools:
    @staticmethod
    def camel_to_snake(camel_case: str) -> str:
        snake_case = []
        for char in camel_case:
            if char.isupper():
                snake_case.append("_" + char.lower())
            else:
                snake_case.append(char)
        return "".join(snake_case).lstrip("_")

    @staticmethod
    def filename_from_url(job_result_url: str) -> str:
        """
        Example of `job_result_url` (str) :
            https://storage.googleapis.com/shopify-tiers-assets-prod-us-east1/<some_hashed_sum>?
                GoogleAccessId=assets-us-prod%40shopify-tiers.iam.gserviceaccount.com&
                Expires=1705508208&
                Signature=<some_long_signature>%3D%3D&
                response-content-disposition=attachment%3B+filename%3D%22bulk-4147374162109.jsonl%22%3B+filename%2A%3DUTF-8%27%27bulk-4147374162109.jsonl&
                response-content-type=application%2Fjsonl

        Output:
            (str): bulk-4147374162109.jsonl
        """
        # Regular expression pattern to extract the filename
        filename_pattern = r'filename\*?=(?:UTF-8\'\')?"([^"]+)"'
        parsed_url = dict(parse_qsl(urlparse(job_result_url).query))
        match = re.search(filename_pattern, parsed_url.get("response-content-disposition", ""))
        if match:
            return match.group(1)
        else:
            raise ShopifyBulkExceptions.BulkJobResultUrlError(
                f"Could not extract the `filename` from `result_url` provided, details: {job_result_url}",
            )

    @staticmethod
    def from_iso8601_to_rfc3339(record: Mapping[str, Any], field: str) -> Mapping[str, Any]:
        """
        Converts date-time as follows:
            Input: "2023-01-01T15:00:00Z"
            Output: "2023-01-01T15:00:00+00:00"
        If the value of the `field` is `None` we return it `as is`.
        """
        # some fields that expected to be resolved as ids, might not be populated for the particular `RECORD`,
        # we should return `None` to make the field `null` in the output as the result of the transformation.
        target_value = record.get(field)
        return pdm.parse(target_value).to_rfc3339_string() if target_value else record.get(field)

    def fields_names_to_snake_case(self, dict_input: Optional[Mapping[str, Any]] = None) -> Optional[MutableMapping[str, Any]]:
        # transforming record field names from camel to snake case, leaving the `__parent_id` relation in place
        if dict_input:
            # the `None` type check is required, to properly handle nested missing entities (return None)
            return {self.camel_to_snake(k) if dict_input and k != BULK_PARENT_KEY else k: v for k, v in dict_input.items()}

    @staticmethod
    def resolve_str_id(
        str_input: Optional[str] = None, output_type: Optional[Union[int, str, float]] = int
    ) -> Optional[Union[int, str, float]]:
        # some fields that expected to be resolved as ids, might not be populated for the particular `RECORD`,
        # we should return `None` to make the field `null` in the output as the result of the transformation.
        if str_input:
            return output_type(re.search(r"\d+", str_input).group())
        else:
            return None



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


from airbyte_cdk.utils import AirbyteTracedException
from airbyte_protocol.models import FailureType


class ShopifyBulkExceptions:
    class BaseBulkException(AirbyteTracedException):
        """Base BULK Job Exception"""

        failure_type: FailureType = FailureType.config_error

        def __init__(self, message: str, **kwargs) -> None:
            super().__init__(internal_message=message, failure_type=self.failure_type, **kwargs)

    class BulkJobError(BaseBulkException):
        """Raised when there are BULK Job Errors in response"""

    class BulkJobUnknownError(BaseBulkException):
        """Raised when BULK Job has FAILED with Unknown status"""

    class BulkJobBadResponse(BaseBulkException):
        """Raised when the requests.Response object could not be parsed"""

    class BulkJobResultUrlError(BaseBulkException):
        """Raised when BULK Job has ACCESS_DENIED status"""

    class BulkRecordProduceError(BaseBulkException):
        """Raised when there are error producing records from BULK Job result"""

    class BulkJobFailed(BaseBulkException):
        """Raised when BULK Job has FAILED status"""

    class BulkJobCanceled(BaseBulkException):
        """Raised when BULK Job has CANCELED status"""

        failure_type: FailureType = FailureType.system_error

    class BulkJobTimout(BaseBulkException):
        """Raised when BULK Job has TIMEOUT status"""

    class BulkJobAccessDenied(BaseBulkException):
        """Raised when BULK Job has ACCESS_DENIED status"""

    class BulkJobConcurrentError(BaseBulkException):
        """Raised when BULK Job could not be created, since the 1 Bulk job / shop quota is exceeded."""



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from typing import Any, Dict, Mapping

from airbyte_cdk.sources.streams.http.requests_native_auth import TokenAuthenticator


class MissingAccessTokenError(Exception):
    """
    Raised when the token is `None` instead of the real value
    """


class NotImplementedAuth(Exception):
    """Not implemented Auth option error"""

    logger = logging.getLogger("airbyte")

    def __init__(self, auth_method: str = None):
        self.message = f"Not implemented Auth method = {auth_method}"
        super().__init__(self.logger.error(self.message))


class ShopifyAuthenticator(TokenAuthenticator):

    """
    Making Authenticator to be able to accept Header-Based authentication.
    """

    def __init__(self, config: Mapping[str, Any]):
        self.config = config

    def get_auth_header(self) -> Mapping[str, Any]:
        auth_header: str = "X-Shopify-Access-Token"
        credentials: Dict = self.config.get("credentials", self.config.get("auth_method"))
        auth_method: str = credentials.get("auth_method")

        if auth_method in ["oauth2.0", "access_token"]:
            access_token = credentials.get("access_token")
            if access_token:
                return {auth_header: access_token}
            else:
                raise MissingAccessTokenError
        elif auth_method == "api_password":
            return {auth_header: credentials.get("api_password")}
        else:
            raise NotImplementedAuth(auth_method)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from dataclasses import dataclass, field
from enum import Enum
from time import sleep, time
from typing import Any, Final, Iterable, List, Mapping, Optional, Union

import pendulum as pdm
import requests
from airbyte_cdk import AirbyteLogger
from requests.exceptions import JSONDecodeError
from source_shopify.utils import ApiTypeEnum
from source_shopify.utils import ShopifyRateLimiter as limiter

from .exceptions import AirbyteTracedException, ShopifyBulkExceptions
from .query import ShopifyBulkTemplates
from .tools import END_OF_FILE, BulkTools


class ShopifyBulkStatus(Enum):
    CREATED = "CREATED"
    CANCELED = "CANCELED"
    CANCELING = "CANCELING"
    COMPLETED = "COMPLETED"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    TIMEOUT = "TIMEOUT"
    ACCESS_DENIED = "ACCESS_DENIED"


@dataclass
class ShopifyBulkManager:
    session: requests.Session
    base_url: str
    stream_name: str

    # default logger
    logger: Final[AirbyteLogger] = logging.getLogger("airbyte")

    # 10Mb chunk size to save the file
    retrieve_chunk_size: Final[int] = 1024 * 1024 * 10
    # time between job status checks
    job_check_interval_sec: Final[int] = 5

    # PLATFORM HEARTBEAT NOTES:
    # 30 sec / attempt * 19 attempts = 570 sec of wait time in total,
    # which is < 10 min of retrying, before Heartbeat will kill the source as non-responsive

    # sleep time per creation attempt
    concurrent_interval_sec: Final[int] = 30
    # max attempts for job creation
    concurrent_max_retry: Final[int] = 19
    # Each job ideally should be executed within the specified time (in sec),
    # to maximize the performance for multi-connection syncs and control the bulk job size within +- 7 mins (420 sec),
    # Ideally the source will balance on it's own rate, based on the time taken to return the data for the slice.
    job_max_elapsed_time_sec: Final[float] = 420.0
    # 0.1 ~= P2H, default value, lower boundary for slice size
    job_size_min: Final[float] = 0.1
    # P365D, upper boundary for slice size
    job_size_max: Final[float] = 365.0
    # running job logger constrain
    log_job_state_msg_frequency: Final[int] = 3
    # attempt limit indicator
    concurrent_max_attempt_reached: bool = field(init=False, default=False)
    # attempt counter
    concurrent_attempt: int = field(init=False, default=0)
    # currents: job_id, job_state, job_created_at, job_self_canceled
    job_id: Optional[str] = field(init=False, default=None)
    job_created_at: Optional[str] = field(init=False, default=None)
    job_self_canceled: bool = field(init=False, default=False)
    job_state: ShopifyBulkStatus = field(init=False, default=None)
    # 2 sec is set as default value to cover the case with the empty-fast-completed jobs
    job_last_elapsed_time: float = field(init=False, default=2.0)
    # dynamically adjusted slice interval
    job_size: float = field(init=False, default=0.0)
    # expand slice factor
    job_size_expand_factor: int = field(init=False, default=2)
    # reduce slice factor
    job_size_reduce_factor: int = field(init=False, default=2)
    # whether or not the slicer should revert the previous start value
    job_should_revert_slice: bool = field(init=False, default=False)
    # running job log counter
    log_job_state_msg_count: int = field(init=False, default=0)

    @property
    def tools(self) -> BulkTools:
        return BulkTools()

    @property
    def job_state_to_fn_map(self) -> Mapping[str, Any]:
        return {
            ShopifyBulkStatus.CREATED.value: self.on_created_job,
            ShopifyBulkStatus.CANCELING.value: self.on_canceling_job,
            ShopifyBulkStatus.CANCELED.value: self.on_canceled_job,
            ShopifyBulkStatus.COMPLETED.value: self.on_completed_job,
            ShopifyBulkStatus.RUNNING.value: self.on_running_job,
            ShopifyBulkStatus.TIMEOUT.value: self.on_timeout_job,
            ShopifyBulkStatus.FAILED.value: self.on_failed_job,
            ShopifyBulkStatus.ACCESS_DENIED.value: self.on_access_denied_job,
        }

    @property
    def job_size_adjusted_expand_factor(self, coef: float = 0.5) -> float:
        """
        The Job Size expand factor is calculated using EMA (Expotentional Moving Average):
            coef - the expantion coefficient
            previous_expand_factor - previous factor value

        Formula: expand_factor = coef * previous_expand_factor + (1 - coef)
        """

        return coef * self.job_size_expand_factor + (1 - coef)

    @property
    def job_size_adjusted_reduce_factor(self) -> float:
        """
        The Job Size reduce factor is 2, by default.
        """

        return self.job_size_reduce_factor

    @property
    def job_elapsed_time_in_state(self) -> int:
        """
        Returns the elapsed time taken while Job is in certain status/state.
        """
        return (pdm.now() - pdm.parse(self.job_created_at)).in_seconds() if self.job_created_at else 0

    @property
    def is_long_running_job(self) -> bool:
        if self.job_elapsed_time_in_state:
            if self.job_elapsed_time_in_state > self.job_max_elapsed_time_sec:
                # set the slicer to revert mode
                self.job_should_revert_slice = True
                return True
        else:
            # reset slicer to normal mode
            self.job_should_revert_slice = False
            return False

    def expand_job_size(self) -> None:
        self.job_size += self.job_size_adjusted_expand_factor

    def reduce_job_size(self) -> None:
        self.job_size /= self.job_size_adjusted_reduce_factor

    def __adjust_job_size(self, job_current_elapsed_time: float) -> None:
        if not self.job_should_revert_slice:
            if job_current_elapsed_time > self.job_max_elapsed_time_sec:
                self.reduce_job_size()
            elif job_current_elapsed_time < 1 or job_current_elapsed_time < self.job_last_elapsed_time:
                self.expand_job_size()
            elif job_current_elapsed_time > self.job_last_elapsed_time < self.job_max_elapsed_time_sec:
                pass

            # set the last job time
            self.job_last_elapsed_time = job_current_elapsed_time
            # check the job size slice interval are acceptable
            self.job_size = max(self.job_size_min, min(self.job_size, self.job_size_max))
        else:
            pass

    def __reset_state(self) -> None:
        # set current job state to default values
        self.job_state, self.job_id, self.job_self_canceled = None, None, False
        # set the running job message counter to default
        self.log_job_state_msg_count = 0

    def job_completed(self) -> bool:
        return self.job_state == ShopifyBulkStatus.COMPLETED.value

    def job_canceled(self) -> bool:
        return self.job_state == ShopifyBulkStatus.CANCELED.value

    def job_cancel(self) -> requests.Response:
        # re-use of `self._session(*, **)` to make BULK Job cancel request
        cancel_args = self.job_get_request_args(ShopifyBulkTemplates.cancel)
        with self.session as cancel_job:
            canceled_response = cancel_job.request(**cancel_args)
            # mark the job was self-canceled
            self.job_self_canceled = True
            return self.job_healthcheck(canceled_response)

    def log_job_state_with_count(self) -> None:
        """
        Print the status/state Job info message every N request, to minimize the noise in the logs.
        """
        if self.log_job_state_msg_count < self.log_job_state_msg_frequency:
            self.log_job_state_msg_count += 1
        else:
            message = f"Elapsed time: {self.job_elapsed_time_in_state} sec"
            self.log_state(message)
            self.log_job_state_msg_count = 0

    def log_state(self, message: Optional[str] = None) -> None:
        pattern = f"Stream: `{self.stream_name}`, the BULK Job: `{self.job_id}` is {self.job_state}."
        if message:
            self.logger.info(f"{pattern}. {message}.")
        else:
            self.logger.info(pattern)

    def job_get_request_args(self, template: ShopifyBulkTemplates) -> Mapping[str, Any]:
        return {
            "method": "POST",
            "url": self.base_url,
            "data": template(self.job_id),
            "headers": {"Content-Type": "application/graphql"},
        }

    def job_get_result(self, response: Optional[requests.Response] = None) -> Optional[str]:
        job_result_url = response.json().get("data", {}).get("node", {}).get("url") if response else None
        if job_result_url:
            # save to local file using chunks to avoid OOM
            filename = self.tools.filename_from_url(job_result_url)
            with self.session.get(job_result_url, stream=True) as response:
                response.raise_for_status()
                with open(filename, "wb") as file:
                    for chunk in response.iter_content(chunk_size=self.retrieve_chunk_size):
                        file.write(chunk)
                    # add `<end_of_file>` line to the bottom  of the saved data for easy parsing
                    file.write(END_OF_FILE.encode())
            return filename

    def job_update_state(self, response: Optional[requests.Response] = None) -> None:
        if response:
            self.job_state = response.json().get("data", {}).get("node", {}).get("status")
            if self.job_state in [ShopifyBulkStatus.RUNNING.value, ShopifyBulkStatus.CANCELING.value]:
                self.log_job_state_with_count()
            else:
                self.log_state()

    def on_created_job(self, **kwargs) -> None:
        pass

    def on_canceled_job(self, response: requests.Response) -> None:
        if not self.job_self_canceled:
            raise ShopifyBulkExceptions.BulkJobCanceled(
                f"The BULK Job: `{self.job_id}` exited with {self.job_state}, details: {response.text}",
            )
        else:
            pass

    def on_canceling_job(self, **kwargs) -> None:
        sleep(self.job_check_interval_sec)

    def on_running_job(self, **kwargs) -> None:
        if self.is_long_running_job:
            self.logger.info(
                f"Stream: `{self.stream_name}` the BULK Job: {self.job_id} runs longer than expected. The job will be canceled and retried with the smaller Slice Size."
            )
            self.job_cancel()
        else:
            sleep(self.job_check_interval_sec)

    def on_completed_job(self, **kwargs) -> None:
        pass

    def on_failed_job(self, response: requests.Response) -> AirbyteTracedException:
        raise ShopifyBulkExceptions.BulkJobFailed(
            f"The BULK Job: `{self.job_id}` exited with {self.job_state}, details: {response.text}",
        )

    def on_timeout_job(self, **kwargs) -> AirbyteTracedException:
        raise ShopifyBulkExceptions.BulkJobTimout(
            f"The BULK Job: `{self.job_id}` exited with {self.job_state}, please reduce the `GraphQL BULK Date Range in Days` in SOURCES > Your Shopify Source > SETTINGS.",
        )

    def on_access_denied_job(self, **kwagrs) -> AirbyteTracedException:
        raise ShopifyBulkExceptions.BulkJobAccessDenied(
            f"The BULK Job: `{self.job_id}` exited with {self.job_state}, please check your PERMISSION to fetch the data for this stream.",
        )

    def on_job_with_errors(self, errors: List[Mapping[str, Any]]) -> AirbyteTracedException:
        raise ShopifyBulkExceptions.BulkJobUnknownError(f"Could not validate the status of the BULK Job `{self.job_id}`. Errors: {errors}.")

    def job_check_for_errors(self, response: requests.Response) -> Iterable[Mapping[str, Any]]:
        try:
            return response.json().get("errors") or response.json().get("data", {}).get("bulkOperationRunQuery", {}).get("userErrors", [])
        except (Exception, JSONDecodeError) as e:
            raise ShopifyBulkExceptions.BulkJobBadResponse(
                f"Couldn't check the `response` for `errors`, status: {response.status_code}, response: `{response.text}`. Trace: {repr(e)}."
            )

    def job_track_running(self) -> Union[AirbyteTracedException, requests.Response]:
        # format Job state check args
        status_args = self.job_get_request_args(ShopifyBulkTemplates.status)
        # re-use of `self._session(*, **)` to make BULK Job status checks
        with self.session as track_running_job:
            response = track_running_job.request(**status_args)
        # errors check
        errors = self.job_check_for_errors(response)
        if not errors:
            self.job_update_state(response)
            self.job_state_to_fn_map.get(self.job_state)(response=response)
            return response
        else:
            # execute ERRORS scenario
            self.on_job_with_errors(errors)

    def job_check_state(self) -> Optional[str]:
        while not self.job_completed():
            response = self.job_track_running()
            if self.job_canceled():
                return None
        # return `job_result_url` when status is `COMPLETED`
        return self.job_get_result(response)

    def has_running_concurrent_job(self, errors: Optional[Iterable[Mapping[str, Any]]] = None) -> bool:
        """
        When concurent BULK Job is already running for the same SHOP we receive:
        Error example:
        [
            {
                'field': None,
                'message': 'A bulk query operation for this app and shop is already in progress: gid://shopify/BulkOperation/4039184154813.',
            }
        ]
        """

        concurent_job_pattern = "A bulk query operation for this app and shop is already in progress"
        # the errors are handled in `job_job_check_for_errors`
        if errors:
            for error in errors:
                message = error.get("message", "") if isinstance(error, dict) else ""
                if concurent_job_pattern in message:
                    return True
        # reset the `concurrent_attempt` counter, once there is no concurrent job error
        self.concurrent_attempt = 0
        return False

    def has_reached_max_concurrency_attempt(self) -> bool:
        return self.concurrent_attempt == self.concurrent_max_retry

    def job_retry_request(self, request: requests.PreparedRequest) -> Optional[requests.Response]:
        with self.session.send(request) as retried_request:
            return retried_request

    def job_retry_concurrent(self, request: requests.PreparedRequest) -> Optional[requests.Response]:
        self.concurrent_attempt += 1
        self.logger.warning(
            f"Stream: `{self.stream_name}`, the BULK concurrency limit has reached. Waiting {self.concurrent_interval_sec} sec before retry, atttempt: {self.concurrent_attempt}.",
        )
        sleep(self.concurrent_interval_sec)
        return self.job_healthcheck(self.job_retry_request(request))

    def job_get_id(self, response: requests.Response) -> Optional[str]:
        response_data = response.json()
        bulk_response = response_data.get("data", {}).get("bulkOperationRunQuery", {}).get("bulkOperation", {})
        if bulk_response and bulk_response.get("status") == ShopifyBulkStatus.CREATED.value:
            job_id = bulk_response.get("id")
            self.job_created_at = bulk_response.get("createdAt")
            self.logger.info(f"Stream: `{self.stream_name}`, the BULK Job: `{job_id}` is {ShopifyBulkStatus.CREATED.value}")
            return job_id
        else:
            return None

    def job_retry_on_concurrency(self, request: requests.PreparedRequest) -> Union[AirbyteTracedException, Optional[requests.Response]]:
        if self.has_reached_max_concurrency_attempt():
            # indicate we're out of attempts to retry with job creation
            message = f"The BULK Job couldn't be created at this time, since another job is running."
            # log the message
            self.logger.error(message)
            # raise AibyteTracebackException with `INCOMPLETE` status
            raise ShopifyBulkExceptions.BulkJobConcurrentError(message)
        else:
            return self.job_retry_concurrent(request)

    def job_healthcheck(self, response: requests.Response) -> Optional[requests.Response]:
        # get the latest request to retry
        request: requests.PreparedRequest = response.request
        try:
            errors = self.job_check_for_errors(response)
            # when the concurrent job takes place, we typically need to wait and retry, but no longer than 10 min.
            if not self.has_running_concurrent_job(errors):
                return response if not errors else None
            else:
                return self.job_retry_on_concurrency(request)
        except (ShopifyBulkExceptions.BulkJobBadResponse, ShopifyBulkExceptions.BulkJobUnknownError) as err:
            # sometimes we face with `HTTP-500 Internal Server Error`
            # we should retry such at least once
            self.logger.info(f"Stream: `{self.stream_name}`, retrying Bad Request: {request.body}, error: {repr(err)}.")
            return self.job_retry_request(request)

    @limiter.balance_rate_limit(api_type=ApiTypeEnum.graphql.value)
    def job_check(self, created_job_response: requests.Response) -> Optional[str]:
        """
        This method checks the status for the BULK Job created, using it's `ID`.
        The time spent for the Job execution is tracked to understand the effort.
        """
        job_response = self.job_healthcheck(created_job_response)
        self.job_id: str = self.job_get_id(job_response)
        job_started = time()
        try:
            return self.job_check_state()
        except (
            ShopifyBulkExceptions.BulkJobCanceled,
            ShopifyBulkExceptions.BulkJobFailed,
            ShopifyBulkExceptions.BulkJobTimout,
            ShopifyBulkExceptions.BulkJobAccessDenied,
            ShopifyBulkExceptions.BulkJobUnknownError,
        ) as bulk_job_error:
            raise bulk_job_error
        finally:
            job_current_elapsed_time = round((time() - job_started), 3)
            self.logger.info(f"Stream: `{self.stream_name}`, the BULK Job: `{self.job_id}` time elapsed: {job_current_elapsed_time} sec.")
            # check whether or not we should expand or reduce the size of the slice
            self.__adjust_job_size(job_current_elapsed_time)
            # reset the state for COMPLETED job
            self.__reset_state()
